<?php
/**
 * @package        AJAX Toggler
 * @copyright      Copyright (C)  2009 - 2023 AlterBrains.com. All rights reserved.
 * @license        GNU/GPL, see LICENSE.txt
 */

/** @noinspection PhpMultipleClassDeclarationsInspection */

namespace AlterBrains\Plugin\System\Ajaxtoggler\Extension;

use Joomla\CMS\Application\AdministratorApplication;
use Joomla\CMS\Application\SiteApplication;
use Joomla\CMS\Environment\Browser;
use Joomla\CMS\Factory;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Response\JsonResponse;
use Joomla\CMS\Toolbar\Toolbar;
use Joomla\CMS\Uri\Uri;
use Joomla\Event\SubscriberInterface;

\defined('_JEXEC') or die;

/**
 * @since 1.0
 */
class Ajaxtoggler extends CMSPlugin implements SubscriberInterface
{
    /**
     * @var SiteApplication|AdministratorApplication
     * @since 1.0
     */
    private $application;

    /**
     * @var bool
     * @since 3.2.0
     */
    protected $active;

    /**
     * @var \Joomla\CMS\Document\HtmlDocument
     * @since 3.2.3
     */
    protected $document;

    /**
     * @var \Joomla\CMS\Session\Session
     * @since 3.2.3
     */
    protected $session;

    /**
     * @inheritDoc
     * @since 4.0
     */
    public function __construct($subject, $config = [])
    {
        parent::__construct($subject, $config);

        $this->application = Factory::getApplication();
    }

    /**
     * @inheritDoc
     * @since 1.0
     */
    public static function getSubscribedEvents(): array
    {
        return [
            'onAfterRoute'        => 'onAfterRoute',
            'onAfterDispatch'     => 'onAfterDispatch',
            'onBeforeCompileHead' => 'onBeforeCompileHead',
        ];
    }

    public function onAfterRoute()
    {
        if (\in_array($this->application->input->get('option'), $this->params->get('exclude', []), true)) {
            return;
        }

        $this->session = $this->application->getSession();

        if ($this->application->input->getBool('jatoggler')) {
            $this->session->set('jatoggler', 1);

            // We should normally redirect in $app->redirect, else we fail with stupid IE
            $navigator = Browser::getInstance();
            if ($navigator->isBrowser('msie')) {
                $navigator->setBrowser('chrome');
            }
        }
    }

    public function onAfterDispatch()
    {
        /** @noinspection PhpUnhandledExceptionInspection */
        $this->document = $this->application->getDocument();

        if (!$this->session || $this->document->getType() !== 'html') {
            return;
        }

        $this->active = true;

        if ($this->session->get('jatoggler')) {
            $this->session->set('jatoggler', 0);

            if (\strtolower($this->application->input->server->get('HTTP_X_REQUESTED_WITH', '')) === 'xmlhttprequest') {
                echo new JsonResponse([
                    'html'       => $this->getTable() . $this->getOrdering(),
                    'toolbar'    => $this->getToolbar(),
                    'pagination' => $this->getPagination(),
                    'title'      => $this->application->JComponentTitle ?? null,
                ]);

                $this->application->close();
            }
        }
    }

    /**
     * @since 3.3.0
     */
    public function onBeforeCompileHead()
    {
        if (!$this->active) {
            return;
        }

        $options = \json_encode([
            'base' => Uri::root(),
        ], JSON_THROW_ON_ERROR);

        /** @noinspection BadExpressionStatementJS */
        /** @noinspection JSVoidFunctionReturnValueUsed */
        $initScript = <<<JS
jQuery(function() {
	AjaxToggler.initialize($options)
});
JS;

        $script = $this->_name . (\JDEBUG ? '.js' : '.min.js');

        $this->document->getWebAssetManager()
            ->registerAndUseScript(
                'plg_' . $this->_name . '.admin',
                'plugins/system/' . $this->_name . '/media/js/' . $script,
                ['version' => 'auto'],
                ['defer' => false],
                ['core', 'jquery']
            )
            ->addInlineScript($initScript)
            ->useScript('webcomponent.core-loader');
    }

    /**
     * @return string
     * @since 1.0
     */
    protected function getTable()
    {
        $html = '';

        // We have table
        // Apply lazy match to capture the full table, use negative lookahead
        \preg_match_all('/(<table((?!<table>)).*?<\/table>)/s', $this->document->getBuffer('component'), $matches);

        if (!empty($matches[1])) {
            foreach ($matches[1] as $table) {
                if (
                    // Should contain all toggle
                    \str_contains($table, '<input class="form-check-input" autocomplete="off"')
                    // We can have multiple tables, use the biggest one
                    && \strlen($html) < \strlen($table)
                    && (
                        // Usual Joomla table
                        \str_contains($table, ' class="table')
                        // K2
                        || \str_contains($table, ' class="adminlist table table-striped')
                    )
                ) {
                    $html = $table;
                }
            }
        }

        // No results! Keep our wrapper ID!
        if ($html === '') {
            $html = '<div id="ajaxtogglerWrapper" class="alert alert-info">' .
                '<span class="icon-info-circle" aria-hidden="true"></span><span class="visually-hidden">' . Text::_('INFO') . '</span> ' .
                Text::_('JGLOBAL_NO_MATCHING_RESULTS') .
                '</div>';
        }

        // small house-keeping
        return $html;
    }

    /**
     * @return string
     * @since 1.0
     */
    protected function getToolbar()
    {
        $html = '';

        // Reload toolbar if Trashed state changed
        $state1 = $this->session->get('jatoggler_filter_published');
        $state2 = $this->application->input->get('filter_published', $this->application->input->get('filter_state'));

        // New Joomla 3.2.1+ filter
        $filter = $this->application->input->get('filter', [], 'array');

        if (isset($filter['published'])) {
            $state2 = $filter['published'];
        } elseif (isset($filter['state'])) {
            $state2 = $filter['state'];
        }

        /** @noinspection TypeUnsafeComparisonInspection */
        if (
            // Reload toolbar if Trashed state changed
            (($state1 == -2 || $state2 == -2) && $state1 != $state2)
            ||
            // Reload toolbar for selected components;
            ($this->params->get('toolbar') && in_array($this->application->input->get('option'), $this->params->get('toolbar', []), true))
        ) {
            $this->session->set('jatoggler_filter_published', $state2);

            // render toolbar
            /** @noinspection PhpDeprecationInspection */
            /** @noinspection PhpRedundantOptionalArgumentInspection */
            $html = Toolbar::getInstance('toolbar')->render();
        }

        return $html;
    }

    /**
     * @return string
     * @since 1.0
     */
    protected function getPagination()
    {
        $html = '';

        $pattern = '/(<ul class="pagination ms-auto mb-4 me-0">.*?<\/ul>)/Us';

        if (\preg_match_all($pattern, $this->document->getBuffer('component'), $matches) && !empty($matches[0][0])) {
            $html = $matches[0][0] . '<input type="hidden" value="' . $this->application->input->getInt('limitstart') . '" name="limitstart" />';
        }

        return $html;
    }

    /**
     * Re-init ordering list if table was ordered by ordering.
     *
     * @return string
     * @since 1.0
     */
    protected function getOrdering()
    {
        // Already loaded
        $exists = $this->application->input->get('jatoggler_ordering');

        $html = '';

        /** @noinspection PhpUnhandledExceptionInspection */
        $ordering = $this->application->input->get('filter_order');

        // New Joomla 3.2.1+ list
        $list = $this->application->input->get('list', [], 'array');

        if (isset($list['fullordering'])) {
            $ordering = \strtr($list['fullordering'], [
                ' ASC'  => '',
                ' DESC' => '',
            ]);
        }

        if (\in_array($ordering, ['a.ordering', 'ordering', 'a.lft', 'lft'], true)) {
            if (!$exists) {
                $html .= $this->extractScript('dragula');
                $html .= $this->extractStylesheet('dragula');
            }

            // Maximum perversion.
            //$html .= $this->extractScript('joomla.draggable');
            if (
                ($item = $this->document->getWebAssetManager()->getAsset('script', 'joomla.draggable'))
                && ($file = JPATH_SITE . '/' . $item->getUri())
                && \is_file($file)
            ) {
                $time = \time();

                // Wrap into function to prevent "Uncaught SyntaxError: Identifier 'url' has already been declared"
                $html .= '<script>'
                    . \strtr('(function () {' . file_get_contents($file) . '}());', ['DOMContentLoaded' => 'DOMContentLoaded' . $time])
                    . '; document.dispatchEvent(new CustomEvent("DOMContentLoaded' . $time . '", {}));'
                    . '</script>';
            }

            // todo, allow custom options. Joomla doesn't add them by default.
            //$opts = $this->document->getScriptOptions('draggable-list');
        }

        return $html;
    }

    /**
     * @param  string  $name
     *
     * @return string|void
     *
     * @since 3.3.0
     */
    protected function extractScript($name)
    {
        if ($item = $this->document->getWebAssetManager()->getAsset('script', $name)) {
            $uri = $item->getUri();
            if ($version = $item->getVersion()) {
                $uri .= (!\str_contains('?', $uri) ? '?' : '&amp;') . $version;
            }

            return '<script src="' . $uri . '"></script>';
        }
    }

    /**
     * @param  string  $name
     *
     * @return string|void
     *
     * @since 3.3.0
     */
    protected function extractStylesheet($name)
    {
        if ($item = $this->document->getWebAssetManager()->getAsset('style', $name)) {
            $uri = $item->getUri();
            if ($version = $item->getVersion()) {
                $uri .= (!\str_contains('?', $uri) ? '?' : '&amp;') . $version;
            }

            return '<link href="' . $uri . '" rel="stylesheet" />';
        }
    }
}
